import geopandas as gpd
import numpy as np
from typing import *
import requests
from io import BytesIO
import zipfile
import time
from scipy.interpolate import RegularGridInterpolator, NearestNDInterpolator

from osm_importer.url_list import URL_LIST
import logging

log = logging.getLogger(__name__)


def get_elevation_from_srtm(tdf: gpd.GeoDataFrame) -> Dict[Any, int]:
	"""
	find elevation value for every points in a GeoDataFrame

	parameters
	----------
	tdf: gpd.GeoDataFrame = gpd to of points in EPSG:4326

	returns
	----------
	{index: elevation} : Dict[Any,int] = elevataion for each index in the GeoDataFrame
	"""

	df = tdf.copy()
	srtm = 'srtm3'  # srtm1 is not working in montreal. only available in the US but the file exist in montreal...
	arc_dict = {'srtm1': 1, 'srtm3': 3}

	# lon is est west
	# lat is north south
	# get lon lat from geometry
	lonlat_list = [list((geom.xy[0][0], geom.xy[1][0])) for geom in df['geometry'].values]
	df[['lon', 'lat']] = lonlat_list

	# get eachPoint file to fetch
	df['file_name'] = df.apply(lambda x: get_file_name(x['lat'], x['lon']), axis=1)

	# read dict [fileName : url] from srtm.py python librairy

	el_dict = {}
	for file in list(df['file_name'].unique()):
		temp_df = df[df['file_name'] == file].copy()

		url = URL_LIST[srtm][file]

		fetch_data(url)

		alt, lat, lon = read_hgt_file('/tmp/{f}'.format(f=file), arc_dict[srtm])

		alt = fill_missing_elevation(alt)

		temp_df['elevation'] = interpolate_elevation(temp_df['lon'].values, temp_df['lat'].values, alt, lon, lat)

		el_dict.update(temp_df['elevation'].to_dict())

	# df['elevation'] = df.index.map(el_dict.get)
	return el_dict


def get_file_name(latitude: float, longitude: float) -> str:
	"""
	from SRTM.py. find srtm file name with lat lon.

	parameters
	----------
	latitude (float) : point latitude (epsg 4326)
	longitude (float) : point longitude (epsg 4326)

	returns
	----------
	filename ex: N45W074.hgt
	"""

	if latitude >= 0:
		north_south = 'N'
	else:
		north_south = 'S'

	if longitude >= 0:
		east_west = 'E'
	else:
		east_west = 'W'

	file_name = '%s%s%s%s.hgt' % (
		north_south,
		str(int(abs(np.floor(latitude)))).zfill(2),
		east_west,
		str(int(abs(np.floor(longitude)))).zfill(3),
	)
	return file_name


def fetch_data(url: str, wd: str = '/tmp') -> None:
	"""
	fetch SRTM zip file and extract it as tmp/<fileName>.hgt

	parameters
	----------
	url (str) : URL of the .zip file to fetch

	returns
	----------
	Nothing. save file in tmp/ folder
	"""

	try:
		response = requests.get(url)
		buffer = BytesIO(response.content)
		with zipfile.ZipFile(buffer, 'r') as zip_ref:
			zip_ref.extractall(wd)
			log.info('file save to', wd)
	except:
		time.sleep(2)
		response = requests.get(url)
		buffer = BytesIO(response.content)
		with zipfile.ZipFile(buffer, 'r') as zip_ref:
			zip_ref.extractall(wd)
			log.info('file save to', wd)


def read_hgt_file(hgt_file: str, arc: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
	"""
	read hgt file, extract the matrix data and the lat lon values

	parameters
	----------
	hgt_file (str) : path of the .hgt file
	arc (int) : resolution on the hgt file. srtm1 : 1, srtm3: 3, etc

	returns
	----------
	elevation (np.array) : matrix (N X N) elevation[lon, lat]
	lat (np.array) : vector  (N x 1)
	lon (np.array) : vector (N X 1)
	"""

	SAMPLES = int(3600 / arc + 1)

	filename = hgt_file.split('/')[-1]

	if filename.split('S')[0] != filename:
		lat0 = -int(filename.split('S')[-1][0:2])
	if filename.split('N')[0] != filename:
		lat0 = int(filename.split('N')[-1][0:2])
	if filename.split('W')[0] != filename:
		lon0 = -int(filename.split('W')[-1][0:3])
	if filename.split('E')[0] != filename:
		lon0 = int(filename.split('E')[-1][0:3])

	lat = np.linspace(lat0, lat0 + 1, SAMPLES)
	lon = np.linspace(lon0, lon0 + 1, SAMPLES)

	with open(hgt_file, 'rb') as hgt_data:
		# Each data is 16bit signed integer(i2) - big endian(>)
		elevation = np.fromfile(hgt_data, np.dtype('>i2'), SAMPLES * SAMPLES).reshape((SAMPLES, SAMPLES))
		return np.flip(elevation, axis=0).T, lat, lon


def fill_missing_elevation(alt: np.ndarray) -> np.ndarray:
	"""
	NaN value are equal to -32768. Interpolated those value with kneighbor

	parameters
	----------
	alt: np.ndarray = Matrix of elevation (from read_hgt_file())

	returns
	----------
	alt: np.ndarray = Matrix of elevation with missing value interpolated from kneighbor
	"""
	mask = np.where(alt != -32768)
	interp = NearestNDInterpolator(np.transpose(mask), alt[mask])
	interp_alt = interp(*np.indices(alt.shape))
	return interp_alt


def interpolate_elevation(
	lon_list: List[float], lat_list: List[float], alt: np.ndarray, lon: np.ndarray, lat: np.ndarray
) -> List[int]:
	"""
	find elevation value at any point with 2d interpolation of SRTM matrix

	parameters
	----------
	lon_list: List[float] = list (or np.ndaray) of longitude to interpolate
	lat_list: List[float] = list (or np.ndaray) of latitude to interpolate
	alt: np.ndarray = Matrix of elevation (from read_hgt_file())
	lon: np.ndarray = vector of longidude for the matrix (from read_hgt_file())
	lat: np.ndarray = vector of latitude for the matrix (from read_hgt_file())

	returns
	----------
	List[int] elevation (meters) at each point values are rounded to int.
	"""

	interp = RegularGridInterpolator((lon, lat), alt, bounds_error=False, fill_value=None)

	return [int(interp(tup)) for tup in zip(lon_list, lat_list)]


def calc_incline(a: np.ndarray, b: np.ndarray, length: np.ndarray) -> np.ndarray:
	"""
	find each road segment Incline (in degs)

	parameters
	----------
	a: np.ndarray[float|int] = array of elevation (node a)
	b: np.ndarray[float|int] = array of elevation (node b)
	length: np.ndarray[float|int] = array of link length (in meters, as elevation)

	returns
	----------
	np.ndarray[float] Incline for each roads.
	"""
	return np.degrees(np.arctan((b - a) / length))



import geopandas as gpd
from typing import *
import logging

log = logging.getLogger(__name__)
CYCLEWAY_COLUMNS = ['cycleway:both', 'cycleway:left', 'cycleway:right']


def simple_bicycle_process(links, highway_list):
	# if tag is non existant. add it
	for tag in CYCLEWAY_COLUMNS:
		if tag not in links.columns:
			links[tag] = 'no'
	links = rename_bicycle_tags(links, 'cycleway')
	links = rename_bicycle_tags(links, 'cycleway:both')
	links = rename_bicycle_tags(links, 'cycleway:left')
	links = rename_bicycle_tags(links, 'cycleway:right')

	links['combine_cycle_tag'] = (
		links['cycleway'] + ' ' + links['cycleway:both'] + ' ' + links['cycleway:left'] + ' ' + links['cycleway:right']
	)

	# simple method. everything with a tag highway is an highway both side. using the road oneway.
	bike_dict = {}
	for string in links['combine_cycle_tag'].unique():
		val = string.split(' ')
		if val[0] == 'yes':
			bike_dict[string] = 'yes'
		elif val[0] == 'shared':
			bike_dict[string] = 'shared'
		elif 'yes' in val[1:]:
			bike_dict[string] = 'yes'
		elif 'shared' in val[1:]:
			bike_dict[string] = 'shared'
		else:
			bike_dict[string] = 'no'
	links['cycleway'] = links['combine_cycle_tag'].apply(lambda x: bike_dict.get(x))
	links.loc[links['highway'] == 'cycleway', 'cycleway'] = 'yes'

	links = links.drop(columns=CYCLEWAY_COLUMNS)
	# remove highway not asked for. (because of cycleway)
	links = links[links['highway'].isin(highway_list)]
	links = links.drop(columns='combine_cycle_tag')
	# links = get_bicycle_oneway(links)
	links['cycleway_reverse'] = links['cycleway']

	return links


def rename_bicycle_tags(links: gpd.GeoDataFrame, col: str, inplace: bool = True) -> gpd.GeoDataFrame:
	"""
	replace tags with no, shared or yes.
	inplace = False will create new column with prefix agg_
	"""
	log.info('original values : ', links[col].unique())
	prefix = '' if inplace else 'agg_'
	links[prefix + col] = links[col]
	# fill NaN with no
	links.loc[links[prefix + col].astype('str').str.lower().isin(['nan', 'none']), prefix + col] = 'no'
	cycle_dict = {'no': 'no', 'shared': 'shared', 'share_busway': 'shared', 'shared_lane': 'shared'}

	links[prefix + col] = links[prefix + col].apply(lambda x: cycle_dict.get(x, 'yes'))
	log.info('rename as : ', list(links[prefix + col].unique()))

	return links


def get_bicycle_oneway(links: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
	"""
	format oneway column for cycleway. oneway = True if there is an inverse link. else False.

	"""
	# get bike lanes
	bike_links = links[links['cycleway'] != 'no'].copy()
	# find links with inverse (or duplicated)
	bike_links['abset'] = [frozenset(el) for el in zip(bike_links['a'], bike_links['b'])]
	abset = bike_links.groupby('abset')[['a']].agg(len)
	oneway_abset = abset[abset['a'] > 1].index.values
	# Oneway = True for all links without inverse (abset > 1)
	bike_links['oneway'] = False
	bike_links.loc[bike_links['abset'].isin(oneway_abset), 'oneway'] = True
	# apply to links
	links.loc[links['cycleway'] != 'no', 'oneway'] = bike_links['oneway']
	return links


def extended_bicycle_process(links):
	links['tags'] = links['tags'].astype(str)
	links['tags'] = links['tags'].apply(lambda x: x.replace("'", '"'))
	links['cycleway'] = links.apply(classify_cycleway_tags, axis=1)
	# if classify_cycleway_tags
	links['cycleway_reverse'] = links['cycleway']
	links.loc[links['cycleway'] == 'Dedicated oneway bike path', 'cycleway_reverse'] = 'No'
	links.loc[links['cycleway'] == 'Dedicated oneway bike path', 'cycleway'] = 'Dedicated bike path'

	# if classify_cycleway_tags_with direction
	# we have tuple(left,right). cycleway is right lane (in country whit right side driving)
	# links['cycleway_reverse'] = links['cycleway_reverse'].apply(lambda x: x[0] if type(x)==tuple  else x )
	# links['cycleway'] = links['cycleway'].apply(lambda x: x[1] if type(x)==tuple  else x )

	return links


# tags classification function
def classify_cycleway_tags(df):
	osmTags = df['tags']
	highwayTag = df['highway']

	if highwayTag == 'cycleway':
		if osmTags.find('"foot": "no"') >= 0 and osmTags.find('"oneway": "yes"') >= 0:
			return 'Dedicated oneway bike path'
		elif osmTags.find('"foot": "no"') >= 0:
			return 'Dedicated bike path'
		else:
			return 'Shared bike path'
	elif highwayTag == 'path':
		if (osmTags.find('"bicycle": "yes"') >= 0 or osmTags.find('"bicycle": "designated"') >= 0) and osmTags.find(
			'"foot": "no"'
		) >= 0:
			return 'Dedicated bike path'
		elif (
			osmTags.find('"bicycle": "yes"') >= 0
			or osmTags.find('"bicycle": "designated"') >= 0
			and osmTags.find('"foot": "yes"') >= 0
		):
			return 'Shared bike path'
		else:
			return 'Pedestrian path/street with cycling not allowed'
	elif highwayTag == 'footway' or highwayTag == 'pedestrian':
		if osmTags.find('"bicycle": "yes"') >= 0 or osmTags.find('"bicycle": "designated"') >= 0:
			return 'Pedestrian path/street with cycling allowed'
		else:
			return 'No'  # "Pedestrian path/street with cycling not allowed"
	elif highwayTag == 'living_street':
		return 'Shared zone'
	elif highwayTag == 'construction':
		if osmTags.find('"cycleway": "track"') >= 0:
			return 'Future cycleway'
		else:
			return 'Future link'
	else:
		if (
			osmTags.find('"cycleway": "track"') >= 0
			or osmTags.find('"cycleway": "opposite_track"') >= 0
			or osmTags.find('"cycleway:left": "track"') >= 0
			or osmTags.find('"cycleway:both": "track"') >= 0
			or osmTags.find('"cycleway:right": "track"') >= 0
		):
			return 'Protected bike lane'

		elif (
			osmTags.find('"cycleway:both": "lane"') >= 0
			and (
				osmTags.find('"cycleway:both:lane": "exclusive"') >= 0
				or osmTags.find('"cycleway:left:lane": "exclusive"') >= 0
			)
			and (osmTags.find('"cycleway:left:buffer:right"') >= 0 or osmTags.find('"cycleway:both:buffer:right"') >= 0)
			and (osmTags.find('"cycleway:left:buffer:left"') == -1 or osmTags.find('"cycleway:both:buffer:left"') == -1)
			or (
				osmTags.find('"cycleway:left": "lane"') >= 0
				and osmTags.find('"cycleway:left:lane": "exclusive"') >= 0
				and osmTags.find('"cycleway:left:buffer:right"') >= 0
				and osmTags.find('"cycleway:left:buffer:left"') == -1
			)
		):
			return 'Buffered bike lane (road-side)'

		elif (
			osmTags.find('"cycleway:both": "lane"') >= 0
			and (
				osmTags.find('"cycleway:both:lane": "exclusive"') >= 0
				or osmTags.find('"cycleway:left:lane": "exclusive"') >= 0
			)
			and (osmTags.find('"cycleway:left:buffer:left"') >= 0 or osmTags.find('"cycleway:both:buffer:left"') >= 0)
			and (
				osmTags.find('"cycleway:left:buffer:right"') == -1 or osmTags.find('"cycleway:both:buffer:right"') == -1
			)
			or (
				osmTags.find('"cycleway:left": "lane"') >= 0
				and osmTags.find('"cycleway:left:lane": "exclusive"') >= 0
				and osmTags.find('"cycleway:left:buffer:left"') >= 0
				and osmTags.find('"cycleway:left:buffer:right"') == -1
			)
		):
			return 'Buffered bike lane (kerb-side)'

		elif (
			osmTags.find('"cycleway:both": "lane"') >= 0
			and (
				osmTags.find('"cycleway:both:lane": "exclusive"') >= 0
				or osmTags.find('"cycleway:left:lane": "exclusive"') >= 0
			)
			and (
				(osmTags.find('"cycleway:left:buffer:both"') >= 0 or osmTags.find('"cycleway:both:buffer:both"') >= 0)
				or (
					osmTags.find('"cycleway:left:buffer:left"') >= 0
					and osmTags.find('"cycleway:left:buffer:right"') >= 0
				)
				or (
					osmTags.find('"cycleway:both:buffer:left"') >= 0
					and osmTags.find('"cycleway:both:buffer:right"') >= 0
				)
			)
			or (
				osmTags.find('"cycleway:left": "lane"') >= 0
				and osmTags.find('"cycleway:left:lane": "exclusive"') >= 0
				and (
					(
						osmTags.find('"cycleway:left:buffer:left"') >= 0
						and osmTags.find('"cycleway:left:buffer:right"') >= 0
					)
					or osmTags.find('"cycleway:left:buffer:both"') >= 0
				)
			)
		):
			return 'Buffered bike lane (both sides)'

		elif (
			osmTags.find('"cycleway": "lane"') >= 0
			or osmTags.find('"cycleway": "opposite_lane"') >= 0
			or osmTags.find('"cycleway:left": "lane"') >= 0
			or osmTags.find('"cycleway:right": "lane"') >= 0
			or osmTags.find('"cycleway:both": "lane"') >= 0
		):
			if (
				(
					osmTags.find('"cycleway:both": "lane"') >= 0
					and (
						osmTags.find('"cycleway:both:lane": "advisory"') >= 0
						or osmTags.find('"cycleway:left:lane": "advisory"') >= 0
					)
				)
				or osmTags.find('"cycleway:left": "lane"') >= 0
				and osmTags.find('"cycleway:left:lane": "advisory"') >= 0
			):
				if (
					osmTags.find('"cycleway:left": "lane"') >= 0
					and osmTags.find('"cycleway:right": "no"') >= 0
					and osmTags.find('"cycleway:left:lane": "advisory"') >= 0
				):
					return 'Advisory bike lane (single side)'
				else:
					return 'Advisory bike lane'
			elif (
				osmTags.find('"cycleway:both:conditional"') >= 0
				or osmTags.find('"cycleway:left:conditional"') >= 0
				or osmTags.find('"cycleway:both:lane:conditional"') >= 0
				or osmTags.find('"cycleway:left:lane:conditional"') >= 0
			):
				return 'Peak hour bike lane'
			else:
				return 'Painted bike lane'
		elif (
			osmTags.find('"cycleway": "shared_lane"') >= 0
			or (
				osmTags.find('"cycleway:both": "shared_lane"') >= 0
				and (
					osmTags.find('"cycleway:both:lane": "pictogram"') >= 0
					or osmTags.find('"cycleway:left:lane": "pictogram"') >= 0
				)
			)
			or (
				osmTags.find('"cycleway:left": "shared_lane"') >= 0
				and osmTags.find('"cycleway:left:lane": "pictogram"') >= 0
			)
			or (
				osmTags.find('"cycleway:right": "shared_lane"') >= 0
				and osmTags.find('"cycleway:right:lane": "pictogram"') >= 0
			)
		):
			return 'Sharrow'
		elif (
			osmTags.find('"cycleway:both": "share_busway"') >= 0
			or osmTags.find('"cycleway:both": "opposite_share_busway"') >= 0
			or osmTags.find('"cycleway:left": "share_busway"') >= 0
			or osmTags.find('"cycleway:left": "opposite_share_busway"') >= 0
		):
			return 'Bus lane with cycling allowed'
		elif osmTags.find('"cycleway:left": "shoulder"') >= 0 or osmTags.find('"cycleway:both": "shoulder"') >= 0:
			return 'Shoulder cyclable'
		elif osmTags.find('"bicycle": "dismount"') >= 0:
			return 'Bicyclists dismount'
		elif osmTags.find('"bicycle": "no"') >= 0:
			return 'No'  # "Cycling prohibited"
		elif osmTags.find('"bike"') >= 0 or osmTags.find('"cycle"') >= 0:
			return 'Possible cycling infrastructure/link'
		else:
			return 'No'  # "Mixed traffic"
		# other


# tags classification function
def classify_cycleway_tags_with_direction(df):
	osmTags = df['tags']
	highwayTag = df['highway']
	res = ['No', 'No']
	if highwayTag == 'cycleway':
		if osmTags.find('"foot": "no"') >= 0 and osmTags.find('"oneway": "yes"') >= 0:
			return ('No', 'Dedicated bike path')
		elif osmTags.find('"foot": "no"') >= 0:
			return 'Dedicated bike path'
		else:
			return 'Shared bike path'
	elif highwayTag == 'path':
		if (osmTags.find('"bicycle": "yes"') >= 0 or osmTags.find('"bicycle": "designated"') >= 0) and osmTags.find(
			'"foot": "no"'
		) >= 0:
			return 'Dedicated bike path'
		elif (
			osmTags.find('"bicycle": "yes"') >= 0
			or osmTags.find('"bicycle": "designated"') >= 0
			and osmTags.find('"foot": "yes"') >= 0
		):
			return 'Shared bike path'
		else:
			return 'Pedestrian path/street with cycling not allowed'
	elif highwayTag == 'footway' or highwayTag == 'pedestrian':
		if osmTags.find('"bicycle": "yes"') >= 0 or osmTags.find('"bicycle": "designated"') >= 0:
			return 'Pedestrian path/street with cycling allowed'
		else:
			return 'No'  # "Pedestrian path/street with cycling not allowed"
	elif highwayTag == 'living_street':
		return 'Shared zone'
	elif highwayTag == 'construction':
		if osmTags.find('"cycleway": "track"') >= 0:
			return 'Future cycleway'
		else:
			return 'Future link'
	else:
		if (
			osmTags.find('"cycleway": "track"') >= 0
			or osmTags.find('"cycleway": "opposite_track"') >= 0
			or osmTags.find('"cycleway:both": "track"') >= 0
			or (osmTags.find('"cycleway:left": "track"') >= 0)
			and (~osmTags.find('"cycleway:left:oneway": "yes"') >= 0)
			or (osmTags.find('"cycleway:right": "track"') >= 0)
			and (~osmTags.find('"cycleway:right:oneway": "yes"') >= 0)
		) and osmTags.find('"oneway:bicycle": "yes"') == -1:
			return 'Protected bike lane'
		else:
			if osmTags.find('"cycleway:left": "track"') >= 0 or osmTags.find('"cycleway:left": "opposite_track"') >= 0:
				res[0] = 'Protected bike lane'
			if (
				osmTags.find('"cycleway:right": "track"') >= 0
				or osmTags.find('"cycleway:right": "opposite_track"') >= 0
			):
				res[1] = 'Protected bike lane'

		if (
			osmTags.find('"cycleway:both": "lane"') >= 0
			and (
				osmTags.find('"cycleway:both:lane": "exclusive"') >= 0
				or osmTags.find('"cycleway:left:lane": "exclusive"') >= 0
			)
			and (osmTags.find('"cycleway:left:buffer:right"') >= 0 or osmTags.find('"cycleway:both:buffer:right"') >= 0)
			and (osmTags.find('"cycleway:left:osm:left"') == -1 or osmTags.find('"cycleway:both:buffer:left"') == -1)
			or (
				osmTags.find('"cycleway:left": "lane"') >= 0
				and osmTags.find('"cycleway:left:lane": "exclusive"') >= 0
				and osmTags.find('"cycleway:left:buffer:right"') >= 0
				and osmTags.find('"cycleway:left:buffer:left"') == -1
			)
		):
			return 'Buffered bike lane (road-side)'

		elif (
			osmTags.find('"cycleway:both": "lane"') >= 0
			and (
				osmTags.find('"cycleway:both:lane": "exclusive"') >= 0
				or osmTags.find('"cycleway:left:lane": "exclusive"') >= 0
			)
			and (osmTags.find('"cycleway:left:buffer:left"') >= 0 or osmTags.find('"cycleway:both:buffer:left"') >= 0)
			and (
				osmTags.find('"cycleway:left:buffer:right"') == -1 or osmTags.find('"cycleway:both:buffer:right"') == -1
			)
			or (
				osmTags.find('"cycleway:left": "lane"') >= 0
				and osmTags.find('"cycleway:left:lane": "exclusive"') >= 0
				and osmTags.find('"cycleway:left:buffer:left"') >= 0
				and osmTags.find('"cycleway:left:buffer:right"') == -1
			)
		):
			return 'Buffered bike lane (kerb-side)'

		elif (
			osmTags.find('"cycleway:both": "lane"') >= 0
			and (
				osmTags.find('"cycleway:both:lane": "exclusive"') >= 0
				or osmTags.find('"cycleway:left:lane": "exclusive"') >= 0
			)
			and (
				(osmTags.find('"cycleway:left:buffer:both"') >= 0 or osmTags.find('"cycleway:both:buffer:both"') >= 0)
				or (
					osmTags.find('"cycleway:left:buffer:left"') >= 0
					and osmTags.find('"cycleway:left:buffer:right"') >= 0
				)
				or (
					osmTags.find('"cycleway:both:buffer:left"') >= 0
					and osmTags.find('"cycleway:both:buffer:right"') >= 0
				)
			)
			or (
				osmTags.find('"cycleway:left": "lane"') >= 0
				and osmTags.find('"cycleway:left:lane": "exclusive"') >= 0
				and (
					(
						osmTags.find('"cycleway:left:buffer:left"') >= 0
						and osmTags.find('"cycleway:left:buffer:right"') >= 0
					)
					or osmTags.find('"cycleway:left:buffer:both"') >= 0
				)
			)
		):
			return 'Buffered bike lane (both sides)'

		elif (
			osmTags.find('"cycleway": "lane"') >= 0
			or osmTags.find('"cycleway": "opposite_lane"') >= 0
			or osmTags.find('"cycleway:left": "lane"') >= 0
			or osmTags.find('"cycleway:right": "lane"') >= 0
			or osmTags.find('"cycleway:both": "lane"') >= 0
		):
			if (
				(
					osmTags.find('"cycleway:both": "lane"') >= 0
					and (
						osmTags.find('"cycleway:both:lane": "advisory"') >= 0
						or osmTags.find('"cycleway:left:lane": "advisory"') >= 0
						or osmTags.find('"cycleway:right:lane": "advisory"') >= 0
					)
				)
				or (
					osmTags.find('"cycleway:left": "lane"') >= 0
					and osmTags.find('"cycleway:left:lane": "advisory"') >= 0
				)
				or (
					osmTags.find('"cycleway:right": "lane"') >= 0
					and osmTags.find('"cycleway:tight:lane": "advisory"') >= 0
				)
			):
				if (
					osmTags.find('"cycleway:left": "lane"') >= 0
					and osmTags.find('"cycleway:right": "no"') >= 0
					and osmTags.find('"cycleway:left:lane": "advisory"') >= 0
				):
					res[0] = 'Advisory bike lane'
				if (
					osmTags.find('"cycleway:right": "lane"') >= 0
					and osmTags.find('"cycleway:left": "no"') >= 0
					and osmTags.find('"cycleway:right:lane": "advisory"') >= 0
				):
					res[1] = 'Advisory bike lane'
				else:
					return 'Advisory bike lane'
			elif (
				osmTags.find('"cycleway:both:conditional"') >= 0
				or osmTags.find('"cycleway:left:conditional"') >= 0
				or osmTags.find('"cycleway:both:lane:conditional"') >= 0
				or osmTags.find('"cycleway:left:lane:conditional"') >= 0
			):
				return 'Peak hour bike lane'
			else:
				if (
					osmTags.find('"cycleway": "lane"') >= 0
					or osmTags.find('"cycleway": "opposite_lane"') >= 0
					or osmTags.find('"cycleway:both": "lane"') >= 0
					or osmTags.find('"oneway:bicycle": "yes"') == -1
				):
					return 'Painted bike lane'
				if osmTags.find('"cycleway:left": "lane"') >= 0:
					res[0] = 'Painted bike lane'
				if osmTags.find('"cycleway:right": "lane"') >= 0:
					res[1] = 'Painted bike lane'
		if osmTags.find('"cycleway": "shared_lane"') >= 0 or (
			osmTags.find('"cycleway:both": "shared_lane"') >= 0
			and osmTags.find('"cycleway:both:lane": "pictogram"') >= 0
		):
			return 'Sharrow'
		else:
			if (
				osmTags.find('"cycleway:left:lane": "pictogram"') >= 0
				or osmTags.find('"cycleway:left": "shared_lane"') >= 0
			):
				res[0] = 'Sharrow'
			if (
				osmTags.find('"cycleway:right:lane": "pictogram"') >= 0
				or osmTags.find('"cycleway:right": "shared_lane"') >= 0
			):
				res[1] = 'Sharrow'

		if (
			osmTags.find('"cycleway:both": "share_busway"') >= 0
			or osmTags.find('"cycleway:both": "opposite_share_busway"') >= 0
			or osmTags.find('"cycleway:left": "share_busway"') >= 0
			or osmTags.find('"cycleway:left": "opposite_share_busway"') >= 0
		):
			return 'Bus lane with cycling allowed'
		elif osmTags.find('"cycleway:left": "shoulder"') >= 0 or osmTags.find('"cycleway:both": "shoulder"') >= 0:
			return 'Shoulder cyclable'
		elif osmTags.find('"bicycle": "dismount"') >= 0:
			return 'Bicyclists dismount'
		elif osmTags.find('"bicycle": "no"') >= 0:
			return 'No'  # "Cycling prohibited"
		elif osmTags.find('"bike"') >= 0 or osmTags.find('"cycle"') >= 0:
			return 'Possible cycling infrastructure/link'
		else:
			return tuple(res)  # "Mixed traffic"
		# other



import requests
import pandas as pd
import geopandas as gpd
import time
from shapely import geometry
from typing import TypeAlias, Optional
import logging

log = logging.getLogger(__name__)
BBOX: TypeAlias = tuple[float, float, float, float]
OVERPASS_URL = 'http://overpass-api.de/api/interpreter'


def _overpass_decorator(query: str):
	"""
	take a overpass proto query get_overpass_query() and add headers and footer
	"""
	header = """
	[out:json][timeout:180];
	(
	"""
	footer = """ 
	);
	out body;
	>;
	out skel qt;
	"""
	return header + query + footer


def get_overpass_query(bbox: BBOX, key: str = 'highway', tag_list: list[str] = []) -> str:
	"""
	bbox : (ymin, xmin, ymax, xmax)
	key: osm key ex:highway will fetch way[highway]
	# tag_list: list of tag to includes ex: way[highway ~ motorway | primary]. if empty import all
	"""
	if len(tag_list) == 0:
		return f'way["{key}"]{bbox};\n'
	else:
		tags = '|'.join([tag for tag in tag_list])
		return f'way["{key}"~"{tags}"]{bbox};\n'


def get_overpass_data(query: str, retries: int = 3) -> dict:

	overpassQuery = _overpass_decorator(query)
	for i in range(1, retries + 1):
		try:
			response = requests.get(OVERPASS_URL, params={'data': overpassQuery}, timeout=60)
			if response.status_code != 200:
				raise Exception(f'Overpass API error {response.status_code}: {response.text[:200]}')

			return response.json()
		except Exception as err:
			log.info(err)
			if i < retries:
				wait = 2**i
				log.info(f'retrying in {wait} seconds')
				time.sleep(wait)
	raise Exception('Could not import data from Overpass API. try again')


# transform


def ways_to_geojson(data: dict, geometry: geometry = geometry.LineString) -> gpd.GeoDataFrame:
	"""
	data: osm fetched data object
	geometry: function: how to create geometry from nodes.
	"""
	ways = pd.DataFrame([d for d in data['elements'] if d['type'] == 'way']).set_index('id')
	nodes = pd.DataFrame([d for d in data['elements'] if d['type'] == 'node']).set_index('id')
	# Convert elements to GeoPandas
	way_exploded = ways.explode('nodes').merge(nodes[['lat', 'lon']], left_on='nodes', right_index=True, how='left')
	geom = way_exploded.groupby('id')[['lon', 'lat']].apply(lambda x: geometry(x.values))
	geom.name = 'geometry'
	ways = gpd.GeoDataFrame(ways.join(geom), crs=4326)

	return ways


def add_tags_as_columns(
	ways: gpd.GeoDataFrame, tags: Optional[list[str]] = None, to_drop: list[str] = ['nodes', 'type']
):
	"""
	create new column with tags key. if None. use all tags
	"""
	tags_df = pd.DataFrame.from_records(ways['tags'].values, index=ways['tags'].index)
	if tags is None:
		cols = tags_df.columns
	else:
		cols = [col for col in tags if col in tags_df.columns]
	ways = ways.drop(columns=to_drop, errors='ignore').join(tags_df[cols])
	return ways.reset_index()


def import_data_from_osm(bbox: BBOX, key: str, geometry: geometry, tag_list: list[str] = []) -> gpd.GeoDataFrame:
	overpassQuery = get_overpass_query(bbox, key, tag_list)
	data = get_overpass_data(overpassQuery)
	ways = ways_to_geojson(data, geometry)
	ways = add_tags_as_columns(ways, tags=[key])
	return ways


def get_bbox(ls: list[list[float]]) -> tuple[float, float, float, float]:
	"""
	from a list of coords [[lon, lat], [lon, lat]], ...] get bbox around

	parameters
	----------
	ls: list of coords

	returns
	----------
	tuple (lat_min, lon_min, lat_max, lon_max)
	"""
	xmin = min([coord[0] for coord in ls])
	xmax = max([coord[0] for coord in ls])
	ymin = min([coord[1] for coord in ls])
	ymax = max([coord[1] for coord in ls])
	return (ymin, xmin, ymax, xmax)



from copy import deepcopy
import geopandas as gpd
from shapely.ops import transform
from shapely.geometry import MultiLineString, Point, LineString
from shapely.ops import linemerge
import numpy as np
import pandas as pd
from itertools import islice
from scipy.sparse.csgraph import dijkstra
import math
from osm_importer.graph_utils import (
	get_edge_path,
	get_nodes_degree,
	sparse_matrix,
	get_path,
	get_strongly_connected_component,
)
from typing import Any, Callable
import logging

log = logging.getLogger(__name__)


def process_list_in_col(col_values: Any, new_type: Any, function: Callable[[Any], Any] = np.nanmean) -> Any:
	"""
	ex: links['maxspeed'].apply(lambda x: process_list_in_col(x,float,np.nanmean)
	apply the function to the list. if its not a list. apply new type.
	"""
	if isinstance(col_values, list):
		return function([new_type(val) for val in col_values])
	else:
		return new_type(col_values)


def remove_list_in_col(col_values: Any, method: str = 'first'):
	"""
	remove list. keep first item if first, last if last. return any other value that are not a list
	"""
	if isinstance(col_values, list):
		if method == 'first':
			return col_values[0]
		else:
			return col_values[-1]
	else:
		return col_values


def get_epsg(lat: float, lon: float) -> int:
	"""
	return EPSG in meter for a given (lat,lon)
	lat is north south
	lon is est west
	"""

	return int(32700 - round((45 + lat) / 90, 0) * 100 + round((183 + lon) / 6, 0))


def get_intersections(gdf):
	count = {}
	for geometry in gdf['geometry'].values:
		for p in geometry.coords:
			count[p] = count.get(tuple(p), 0) + 1
	return {k for k, v in count.items() if v > 1}


def split_feature(geom, nodes=(), start=0):
	length = len(geom)
	if length == 2:
		return [geom]
	if length > 2:
		cut = 0
		for p in geom[start + 1 : length - 1]:
			cut += 1
			if tuple(p) in nodes:
				left = geom
				right = geom
				left = left[start : cut + 1]
				right = right[cut:]
				return [left] + split_feature(right, nodes=nodes, start=0)
	return [geom]


def get_split_features(geojson_dict):
	features = []
	intersections = get_intersections(geojson_dict)
	for feature in geojson_dict['features']:
		features += split_feature(feature, nodes=intersections)
	return features


def split_features(gdf):
	intersections = get_intersections(gdf)
	gdf['coords'] = gdf['geometry'].apply(lambda x: x.coords)
	df = gdf.drop(columns=['geometry'])
	df['coords'] = df['coords'].apply(lambda x: split_feature(x, intersections))
	df = df.explode('coords')
	geometry = df['coords'].apply(LineString)

	return gpd.GeoDataFrame(df.drop(columns=['coords']), geometry=geometry, crs=gdf.crs)


def split_directions(geojson_dict):
	features = geojson_dict['features']
	reversed_features = [deepcopy(f) for f in features if not f['properties']['oneway']]
	for f in reversed_features:
		f['geometry']['coordinates'] = list(reversed(f['geometry']['coordinates']))
	geojson_dict['features'] = features + reversed_features


def get_links_and_nodes(links: gpd.GeoDataFrame) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]:
	"""
	read overpass way.geojson file and return links and nodes as gpd.GeoDataFrames.
	Split direction dangerous here as Oneway tag is still rough (not just true / false...)

	parameters
	----------
	geojson_file (str) : path of the file
	split_direction (bool) : if we want to duplicate oneway=False and reverse the geom.

	returns
	----------
	links, nodes
	"""
	links = split_features(links)

	# get nodes
	links['a'] = links['geometry'].apply(lambda x: x.coords[0])
	links['b'] = links['geometry'].apply(lambda x: x.coords[-1])

	node_coordinates = set(links['a']).union(links['b'])
	node_index = {coord: i for i, coord in enumerate(node_coordinates)}
	nodes = gpd.GeoDataFrame(geometry=[Point(pt) for pt in node_coordinates], crs=links.crs)
	nodes.index.name = 'index'

	links['a'] = links['a'].apply(node_index.get)
	links['b'] = links['b'].apply(node_index.get)

	if links.index.name == 'id':
		links = links.reset_index()
	else:
		links = links.reset_index(drop=True)
	links = links.rename(columns={'id': 'osmid'})
	links.index.name = 'index'

	return links, nodes


def main_strongly_connected_component(
	links: gpd.GeoDataFrame, nodes: gpd.GeoDataFrame = None, split_direction=True
) -> gpd.GeoDataFrame:
	"""
	use split direction if the links are not duplicated (if there is 1 link for oneway=False)
	remove cul-de-sac
	"""
	edges = links[['a', 'b']].values
	if 'oneway' in links.columns and split_direction:
		reversed_edges = links.loc[~links['oneway'].astype(bool)][['b', 'a']].values.tolist()
		edges = np.concat([edges, reversed_edges])
	main_scc = set(get_strongly_connected_component(edges))
	_links = links[links['a'].isin(main_scc) & links['b'].isin(main_scc)]
	if nodes is not None:
		_nodes = nodes.loc[list(main_scc)]
		return _links, _nodes
	return _links


def reverse_geom(geom):
	def _reverse(x, y, z=None):
		if z:
			return x[::-1], y[::-1], z[::-1]
		return x[::-1], y[::-1]

	return transform(_reverse, geom)


def rectify_geometry_direction(links, nodes):
	node_dict = nodes['geometry'].to_dict()
	links['node_geom'] = links['a'].apply(lambda x: node_dict.get(x))
	sliced = links['geometry'].apply(lambda x: x.coords[0]) != links['node_geom'].apply(lambda x: x.coords[0])
	log.info(len(links[sliced]), 'geometry to inverse')
	links.loc[sliced, 'geometry'] = links.loc[sliced]['geometry'].apply(lambda x: reverse_geom(x))
	links = links.drop(columns='node_geom')
	return links


def batched(iterable, n):
	"Batch data into lists of length n. The last batch may be shorter."
	# batched('ABCDEFG', 3) --> ABC DEF G
	it = iter(iterable)
	while True:
		batch = list(islice(it, n))
		if not batch:
			return
		yield batch


def cut(line, distance, normalized=True):
	if normalized:
		distance *= line.length

	# Cuts a line in two at a distance from its starting point
	if distance <= 0.0:
		return [None, LineString(line)]
	if distance >= line.length:
		return [LineString(line), None]
	coords = list(line.coords)
	pd = 0
	for i, p in enumerate(coords):
		if i == 0:
			continue
		pd += math.dist(p, coords[i - 1])
		if pd == distance:
			return [LineString(coords[: i + 1]), LineString(coords[i:])]
		if pd > distance:
			cp = line.interpolate(distance)
			return [LineString(coords[:i] + [(cp.x, cp.y)]), LineString([(cp.x, cp.y)] + coords[i:])]


def simplify(links, cutoff=10):
	# create a graph and find all nodes with deg >2 (sources)
	deg_dict = get_nodes_degree(links[['a', 'b']].values)
	sources = [node for node, degree in deg_dict.items() if degree > 2]
	links['weight'] = 1
	log.info(len(sources), 'deg 2 nodes')
	path_to_merge = []
	# graph on oneway and highway as unique as we dont want to aggregate highway together or one way
	for col1, col2 in [(a, b) for a in links['oneway'].unique() for b in links['highway'].unique()]:
		filtered_links = links[(links['oneway'] == col1) & (links['highway'] == col2)]
		if len(filtered_links) < 2:
			continue
		nodes_set = set(filtered_links['a']).union(set(filtered_links['b']))
		mat, node_index = sparse_matrix(filtered_links[['a', 'b', 'weight']].values)
		sparse_deg_dict = {node_index[key]: val for key, val in deg_dict.items() if key in nodes_set}
		index_node = {v: k for k, v in node_index.items()}
		filtered_sources = [s for s in sources if s in nodes_set]
		unfounds_origins = []

		# 1) for each nodes with deg > 2
		# 2) take all non empty path
		# 3) take all path with destination a node with deg > 2
		#    and only deg 2 nodes in between ex: [3,2,2,4]
		# 4) take all path with destination a node of deg == 2
		#    if the path length is the cutoff.

		# batch the dijkstra, my computer was crashing with ~40 000 sources
		# it it not that much slower.
		for origins in batched(filtered_sources, 1000):
			origin_sparse = [node_index[x] for x in origins]
			dist_matrix, predecessors = dijkstra(
				csgraph=mat, directed=True, indices=origin_sparse, return_predecessors=True, limit=cutoff
			)

			# from all source (source are deg>2 by construction)
			for oi in range(len(origins)):
				# destinations are non inf paths
				destinations = np.where(dist_matrix[oi, :] != np.inf)[0]
				# for each destination with deg >= 2
				for di in (d for d in destinations if sparse_deg_dict[d] >= 2):
					path = get_path(predecessors, oi, di)  # get path, with origin and destinations in the list
					path_deg = [*map(sparse_deg_dict.get, path)]  # get deg of each nodes in path
					# if destination is not 2. keep path if every nodes in between are 2
					if path_deg[-1] != 2:
						if set(path_deg[1:-1]) == {2}:
							path_to_merge.append([*map(index_node.get, path)])
							if len(path_to_merge) > 1:
								if path_to_merge[-1] == path_to_merge[-2]:
									log.info(path)
					# if destination == 2
					elif len(path) == cutoff + 1:
						# keep path if len(path) == cutoff+1. at the end redo on those with large cutoff
						if set(path_deg[1:]) == {2}:
							unfounds_origins.append(origins[oi])

		# redo with large cutoff the path thats were longer than the cutoff
		# ex:cutoff = 5, path = [3,2,2,2,2]. the real path may be [3,2,2,2,2,2,2,2,4]. and we will find it now.
		if len(unfounds_origins) > 0:
			log.info('find path with large cutoff for', len(unfounds_origins), ' origins')
			origin_sparse = [node_index[x] for x in unfounds_origins]
			dist_matrix, predecessors = dijkstra(
				csgraph=mat, directed=True, indices=origin_sparse, return_predecessors=True, limit=100
			)
			for oi in range(len(unfounds_origins)):
				# destinations are non inf paths
				destinations = np.where(dist_matrix[oi, :] != np.inf)[0]
				# for each destination with deg >= 2
				for di in (d for d in destinations if sparse_deg_dict[d] >= 2):
					path = get_path(predecessors, oi, di)  # get path, with origin and destinations in the list
					path_deg = [*map(sparse_deg_dict.get, path)]  # get deg of each nodes in path
					# if destination is not 2. keep path if every nodes in between are 2
					# those were not find in cutoff, only check for larger cutoff.
					if (path_deg[-1] != 2) & (len(path) >= cutoff + 1):
						if set(path_deg[1:-1]) == {2}:
							path_to_merge.append([*map(index_node.get, path)])

	# transform sparse nodes path to links path ([rlinks_2, rlinks_40, ...])
	links_dict = links.reset_index().set_index(['a', 'b'])['index'].to_dict()
	links_paths = [[*map(links_dict.get, get_edge_path(path))] for path in path_to_merge]

	# apply a group number to each sequence to merge. group 0 is nothing to merge.
	links['group'] = 0
	idx = 1
	for path in links_paths:
		links.loc[path, 'group'] = idx
		idx += 1
	tlinks = links[links['group'] > 0].copy()
	# sort tlinks links to merge in the right order
	flat_index = [l for sublist in links_paths for l in sublist]
	tlinks = tlinks.loc[flat_index]

	def list_or_first(x: pd.Series):
		return list(x) if x.nunique() > 1 else x.iloc[0]

	def merge_lines(x):
		return linemerge(MultiLineString(list(x)))

	# merge links
	agg_dict = {col: list_or_first for col in tlinks.columns}
	agg_dict['geometry'] = merge_lines
	agg_dict['a'] = 'first'
	agg_dict['b'] = 'last'
	agg_dict['index'] = 'first'
	tlinks = tlinks.reset_index()

	tlinks = tlinks.groupby('group').agg(agg_dict)
	tlinks = tlinks.set_index('index')

	# remove agg that change oneway
	log.info(
		len(tlinks[tlinks['oneway'].apply(lambda x: type(x) == list)]),
		'links were not merge because the oneway field is not the same',
	)
	tlinks = tlinks[~tlinks['oneway'].apply(lambda x: type(x) == list)]

	log.info(
		len(tlinks[tlinks['highway'].apply(lambda x: type(x) == list)]),
		'links were not merge because the highway field is not the same',
	)
	tlinks = tlinks[~tlinks['highway'].apply(lambda x: type(x) == list)]

	# TODO: on a des multiLinestring. il faudrait mieux merger.
	log.info(
		len(tlinks[tlinks['geometry'].apply(lambda x: x.geom_type != 'LineString')]),
		'merged_links unmerged because the geometry became a multilinestring',
	)
	tlinks = tlinks[tlinks['geometry'].apply(lambda x: x.geom_type == 'LineString')]

	# merge tlinks back into links
	tlinks = gpd.GeoDataFrame(tlinks, crs=links.crs)
	links = links[~links['group'].isin(tlinks['group'].unique())]
	links = pd.concat([links, tlinks])
	links = links.drop(columns=['weight', 'group'])

	return links


def split_oneway(links: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
	"""
	duplicated all links with oneway == False. reverse there geometry, reverse a,b and rename with _r
	"""
	nlinks = links[~links['oneway']].copy()
	nlinks.index = nlinks.index.astype(str)  # index must be string
	nlinks.index = nlinks.index + '_r'
	nlinks['geometry'] = nlinks['geometry'].apply(lambda x: reverse_geom(x))
	nlinks = nlinks.rename(columns={'a': 'b', 'b': 'a'})
	links = pd.concat([links, nlinks])
	links['oneway'] = True
	return links


def split_duplicated_links(
	links: gpd.GeoDataFrame, nodes: gpd.GeoDataFrame
) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]:
	crs = nodes.crs
	# index must be string
	links.index = links.index.astype(str)
	nodes.index = nodes.index.astype(str)
	links['a'] = links['a'].apply(str)
	links['b'] = links['b'].apply(str)

	links = links.reset_index()
	duplicated_links = links[links.duplicated(subset=['a', 'b'])].copy()
	log.info(f'Splitting {len(duplicated_links)} duplicated links')
	duplicated_links['geometries'] = duplicated_links['geometry'].apply(cut, distance=0.5)
	duplicated_links['index'] = duplicated_links['index'].apply(lambda x: [x + '_a', x + '_b'])
	middle_nodes = duplicated_links['geometries'].apply(lambda g: Point(g[0].coords[-1]))
	middle_nodes.index = duplicated_links['a'] + '_m'
	middle_nodes.name = 'geometry'

	duplicated_links['middle_node'] = middle_nodes.index
	duplicated_links['a'] = duplicated_links[['a', 'middle_node']].values.tolist()
	duplicated_links['b'] = duplicated_links[['middle_node', 'b']].values.tolist()

	cols = ['index', 'a', 'b', 'geometries']
	other_cols = [c for c in duplicated_links.columns if c not in cols]
	duplicated_links = duplicated_links.set_index(other_cols).apply(pd.Series.explode).reset_index()
	duplicated_links = duplicated_links.drop(columns=['geometry', 'middle_node']).rename(
		columns={'geometries': 'geometry'}
	)

	links = pd.concat([links[~links.duplicated(subset=['a', 'b'])].copy(), duplicated_links])
	links = links.set_crs(crs, allow_override=True)
	nodes = pd.concat([nodes, middle_nodes])
	nodes = nodes.set_crs(crs, allow_override=True)

	return links.set_index('index'), gpd.GeoDataFrame(nodes, crs=crs)


def drop_duplicated_links(
	links: gpd.GeoDataFrame, sort_column: str = 'maxspeed', ascending: bool = False
) -> gpd.GeoDataFrame:
	"""
	drop duplicated links (a,b) with condition sort_column. if maxspeed and ascending=False, keep faster road
	"""
	before = len(links)
	links = links.sort_values(sort_column, ascending=ascending).drop_duplicates(subset=['a', 'b']).sort_index()
	log.info(before - len(links), 'links dropped')
	return links


def clean_maxspeed(links: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
	"""
	remove kph, remove and convert mph. transport everything to float. NaN if string (ex: walk)
	"""
	links['maxspeed'] = links['maxspeed'].str.lower().str.replace('kph', '')
	try:
		mph_index = links['maxspeed'].astype(str).str.lower().str.contains('mph')
		links.loc[mph_index, 'maxspeed'] = (
			links.loc[mph_index, 'maxspeed'].str.lower().str.replace('mph', '').astype('float') * 1.60934
		)
	except:
		log.info('fail to convert mph in maxspeed to float ')
	# this remove all strings lefts
	links['maxspeed'] = links['maxspeed'].apply(pd.to_numeric, errors='coerce')
	return links


def clean_oneway(links: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
	"""
	clean oneways. return True or False
	"""
	# only yes is oneway. all other tags set to false  'no' '-1', 'alternating' 'reversible'
	oneway_dict = {'yes': True}
	links['oneway'] = links['oneway'].apply(lambda x: oneway_dict.get(x, False))
	if len(links['oneway'].unique()) > 2:
		log.info('WARNING: some oneway tags are not defined', links.unique())
	return links


def clean_lanes(links: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
	"""
	remove semicolon (ex lanes = 1;2) and average the result (1;2 -> 1.5).
	transform everything to float. NaN if string
	"""
	semicolon_index = links['lanes'].astype(str).str.contains(';')
	try:
		links.loc[semicolon_index, 'lanes'] = links.loc[semicolon_index, 'lanes'].apply(
			lambda x: np.mean([float(el) for el in x.split(';')])
		)
	except:
		pass
	links['lanes'] = links['lanes'].apply(pd.to_numeric, errors='coerce')
	return links


def fill_na_col(
	links: gpd.GeoDataFrame, group: str = 'highway', col: str = 'maxspeed', agg_func: Callable[[list], Any] = np.mean
) -> gpd.GeoDataFrame:
	"""
	fill the nan value with the average of its group (if np.mean)
	for example: with group=highway, col=maxspeed and agg_func = np.mean. the mean maxpeed per highway is compute.
	then its apply to every NaN value of each highway.
	Note that a group with only NaN will stay a nan (no agg_func will be applied.)

	parameters
	----------
	links: gpd.GeoDataFrame = links to apply the function
	group: group to do the groupby and find the average (or any other function)
	col: column to fill
	agg_func : function to apply

	returns
	----------
	{index: elevation} : Dict[Any,int] = elevataion for each index in the GeoDataFrame
	"""

	val_dict = links[[group, col]].dropna().groupby(group)[col].agg(agg_func).to_dict()
	links.loc[~np.isfinite(links[col]), col] = links.loc[~np.isfinite(links[col]), group].apply(
		lambda x: val_dict.get(x)
	)
	return links


def get_columns_with_list(df: pd.DataFrame) -> list:
	"""
	return the list of column that contain list in them.
	"""

	def _contain_list(df, col):
		for val in df[col].values:
			if type(val) == list:
				return True
		else:
			return False

	object_columns = df.columns[(df.dtypes == object).values]
	res = []
	for col in object_columns:
		if _contain_list(df, col):
			res.append(col)
	return res



import numpy as np
import logging
from osm_importer.road import (
	get_links_and_nodes,
	clean_oneway,
	clean_maxspeed,
	clean_lanes,
	rectify_geometry_direction,
	drop_duplicated_links,
	simplify,
	split_oneway,
	main_strongly_connected_component,
	process_list_in_col,
	remove_list_in_col,
	fill_na_col,
	get_epsg,
	split_duplicated_links,
)
from osm_importer.overpass import get_overpass_query, get_overpass_data, add_tags_as_columns, ways_to_geojson
from osm_importer.bike import simple_bicycle_process, extended_bicycle_process
from osm_importer.elevation import get_elevation_from_srtm, calc_incline
import geopandas as gpd
from typing import Optional
from shapely.geometry import LineString


log = logging.getLogger(__name__)


def import_road_network(
	bbox: tuple[float, float, float, float],
	highway_list: list[str],
	cycleway_list: Optional[list[str]] = None,
	tags: Optional[list[str]] = ['highway', 'maxspeed', 'lanes', 'name', 'oneway', 'surface'],
) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]:
	"""
	add cycleway to tags if you have cycleways. (ex: ['highway', 'maxspeed', 'lanes', 'name', 'oneway', 'surface','cycleway'])
	ex of cycleway_list : ['lane','opposite','opposite_lane','track','opposite_track','share_busway','opposite_share_busway','shared_lane',]
	"""

	# create Query
	query = get_overpass_query(bbox, key='highway', tag_list=highway_list)
	if cycleway_list is not None:
		cycleway_query = get_overpass_query(bbox, key='cycleway', tag_list=cycleway_list)
		query += cycleway_query
	# get data
	log.info('Query road network ...')
	data = get_overpass_data(query, retries=3)
	# transform to geojson
	df = ways_to_geojson(data, LineString)
	df = add_tags_as_columns(df, tags=tags)
	# Create links and nodes netowrks from ways of OSM
	log.info('Convert ways to links and node ...')
	links, nodes = get_links_and_nodes(df)
	links = links.drop(columns='osmid')

	return links, nodes


def simplify_bicycle_network(
	links: gpd.GeoDataFrame, highway_list: list[str], extended_cycleway: bool = False
) -> gpd.GeoDataFrame:
	"""
	return links with processed cycleways tags
	"""
	if 'cycleway' in links.columns:
		if extended_cycleway:
			links = extended_bicycle_process(links)
		else:
			links = simple_bicycle_process(links, highway_list)

	return links


def simplify_road_network(
	links: gpd.GeoDataFrame,
	nodes: gpd.GeoDataFrame,
	add_elevation: bool = True,
	split_direction: bool = False,
	keep_detour: bool = False,
):

	# simplify Linestring geometry. (remove anchor nodes)
	links.geometry = links.simplify(0.00005)

	links = links.drop(columns='tags', errors='ignore')
	# convert oneway to bool.
	links = clean_oneway(links)

	# remove string in maxspeed
	links = clean_maxspeed(links)

	# remove string in maxspeed
	links = clean_lanes(links)

	# make sure the geometry are in the right direction (a->b)
	links = rectify_geometry_direction(links, nodes)

	# remove duplicated links (a-b)
	log.info('simplifying links ...')
	if keep_detour:
		links, nodes = split_duplicated_links(links, nodes)
	else:
		links = drop_duplicated_links(links)

	# simplify. remove deg 2 nodes when possible. group by oneway and highway to merge each links.
	links = simplify(links)

	# split onwway into 2 links a-b, b-a
	if split_direction:
		links = split_oneway(links)

	# Clean Cul de Sac
	log.info('Remove Cul de Sac ...')
	links, nodes = main_strongly_connected_component(links, nodes, not split_direction)

	log.info('removing list in columns ...')
	links['maxspeed'] = links['maxspeed'].apply(lambda x: process_list_in_col(x, float, np.nanmean))
	links['lanes'] = links['lanes'].apply(lambda x: process_list_in_col(x, float, lambda x: np.floor(np.nanmean(x))))
	if 'cycleway' in links.columns:
		# sort and take last. sorted = [no,shared,yes]. so yes or shared if there is a list
		links['cycleway'] = links['cycleway'].apply(lambda x: process_list_in_col(x, str, lambda x: np.sort(x)[-1]))
		if 'cycleway_reverse' in links.columns:
			links['cycleway_reverse'] = links['cycleway_reverse'].apply(
				lambda x: process_list_in_col(x, str, lambda x: np.sort(x)[-1])
			)

	for col in ['highway', 'name', 'surface']:
		links[col] = links[col].apply(lambda x: remove_list_in_col(x, 'first'))

	# Fill NaN with mean values by highway
	links = fill_na_col(links, 'highway', 'maxspeed', lambda x: np.mean(x))
	links = fill_na_col(links, 'highway', 'lanes', lambda x: np.floor(np.mean(x)))

	# Add length
	epsg = get_epsg(nodes.iloc[0]['geometry'].y, nodes.iloc[0]['geometry'].x)
	links['length'] = links.to_crs(epsg).length

	# Add Time
	links['time'] = links['length'] / (links['maxspeed'] * 1000 / 3600)
	links = links.rename(columns={'maxspeed': 'speed'})

	# reindex and remove ununsed nodes
	links = links.reset_index(drop=True)
	links.index = 'rlink_' + links.index.astype(str)
	nodes_set = set(links['a']).union(set(links['b']))
	nodes = nodes.loc[list(nodes_set)].sort_index()

	if add_elevation:
		log.info('Adding elevation')
		el_dict = get_elevation_from_srtm(nodes)
		nodes['elevation'] = nodes.index.map(el_dict.get)
		# incline from node a to b in deg. neg if going down (if b is lower dans a)
		links['incline'] = calc_incline(
			links['a'].apply(lambda x: el_dict.get(x)).values,
			links['b'].apply(lambda x: el_dict.get(x)).values,
			links['length'].values,
		)

	return links, nodes



import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import connected_components
from numba import jit
import numba as nb
from typing import Any


# buildindex
def build_index(edges):
	nodelist = sorted({e[0] for e in edges}.union({e[1] for e in edges}))
	nlen = len(nodelist)
	return dict(zip(nodelist, range(nlen)))


def sparse_matrix(edges, index=None):
	if index is None:
		index = build_index(edges)
	nlen = len(index)
	coefficients = zip(*((index[u], index[v], w) for u, v, w in edges))
	row, col, data = coefficients
	return csr_matrix((data, (row, col)), shape=(nlen, nlen)), index


def get_nodes_degree(edges) -> dict[Any, int]:
	"""
	take edges like links[['a','b]].values
	return node degree
	"""
	node_index = build_index(edges)
	index_node = {v: k for k, v in node_index.items()}
	nlen = len(node_index)
	row = np.array([*map(node_index.get, [e[0] for e in edges])], dtype=np.int32)
	col = np.array([*map(node_index.get, [e[1] for e in edges])], dtype=np.int32)
	data = np.ones(len(edges), dtype=bool)
	mat = csr_matrix((data, (row, col)), shape=(nlen, nlen))

	out_degree = np.array(mat.sum(axis=1)).ravel()  # to 1d arr
	in_degree = np.array(mat.sum(axis=0)).ravel()  # to 1d arr

	degree = in_degree + out_degree
	degree_dict = {index_node.get(i): d for i, d in enumerate(degree)}
	return degree_dict


@jit(nopython=True, locals={'predecessors': nb.int32[:, ::1], 'i': nb.int32, 'j': nb.int32})
def get_path(predecessors, i, j):
	path = [j]
	k = j
	p = 0
	while p != -9999:
		k = p = predecessors[i, k]
		path.append(p)
	return path[::-1][1:]


def get_edge_path(p):
	return list(zip(p[:-1], p[1:]))


def get_strongly_connected_component(edges) -> list[Any]:
	"""
	take edges like links[['a','b]].values
	return the nodes that are in the main connected_componant (biggest)
	"""
	node_index = build_index(edges)
	index_node = {v: k for k, v in node_index.items()}
	nlen = len(node_index)
	# create a directed graph
	row = np.array([*map(node_index.get, [e[0] for e in edges])], dtype=np.int32)
	col = np.array([*map(node_index.get, [e[1] for e in edges])], dtype=np.int32)
	data = np.ones(len(edges), dtype=bool)
	mat = csr_matrix((data, (row, col)), shape=(nlen, nlen))
	# get connected components
	_, labels = connected_components(mat, directed=True, connection='strong')
	# get main component.
	counts = np.bincount(labels)
	main_label = np.where(counts == counts.max())[0][0]
	# return only main component nodes
	main_nodes = [index_node.get(i) for i, d in enumerate(labels) if d == main_label]
	return main_nodes
